body {
    font-family : "Open Sans", sans-serif;
    
    padding-left: 16px;
    padding-right: 16px;
    
    line-height: 1.3;
    text-align: justify;
}

@media screen {
    body {
        /* Bodge for certain mobile browsers which show stuff across the bottom
         * of the screen preventing you reading the last few lines of a recipe.
        */
        padding-bottom: 64px;
    }
}

/* Lay out the page within an (up-to) 20cm central column, with the exception
 * of recipe blocks which may exceed this width up to the screen width as
 * required.
 */
body {
    display: flex;
    flex-direction: column;
    align-items: center;
}

body > * {
    width: 100%;
    max-width: 20cm;
}

body > .rg-recipe-block {
    width: auto;
    max-width: none;
    margin-left: auto;
    margin-right: auto;
}

/* Fill printed pages, don't make artificially narrower */
@media print {
    body {
        padding: 0;
    }
    
    body > *{
        max-width: none;
    }
}

h1, h2, h3, h4, h5, h6 {
    font-family : "Donegal One", "DejaVu Serif", "Times New Roman", Times, serif;
    font-weight: normal;
}

p {
    margin-top: 16px;
    margin-bottom: 16px;
}

header {
    text-align: center;
    
    border-bottom-style: solid;
    border-bottom-width: 1px;
    border-bottom-color: var(--hr-colour);
    
    padding-bottom: 16px;
    margin-bottom: 16px;
}

header h1 {
    margin-bottom: 0;
}
header p {
    margin-top: 8px;
    margin-bottom: 0;
    
    font-style: italic;
    font-size: 90%;
    color: var(--subtitle-colour);
}

a {
    text-decoration: none;
    color: var(--link-colour);
}
a:hover {
    text-decoration: underline;
}
