<!doctype html>
<html lang="en">
    <head>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>{% block title %}{{ title }}{% endblock %}</title>
        {% block stylesheet %}<link rel="stylesheet" href="{{ css_href }}">{% endblock %}
    </head>
    <body>
        {% block nav %}{% endblock %}
        {% block body %}{% endblock %}
    </body>
</html>
